// SPDX-License-Identifier: Apache-2.0
// Copyright Contributors to the OpenTimelineIO project

#pragma once

#include "any/any.hpp"
#include "opentimelineio/version.h"

namespace opentimelineio { namespace OPENTIMELINEIO_VERSION {

using linb::any;
using linb::any_cast;

}} // namespace opentimelineio::OPENTIMELINEIO_VERSION
