#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unsupported contrib code for OpenTimelineIO."""

# flake8: noqa

from . import (
    adapters
)

__version__ = "0.15.0"
__author__ = "Contributors to the OpenTimelineIO project"
__author_email__ = "otio-discussion@lists.aswf.io"
__license__ = "Modified Apache 2.0 License"
