// SPDX-License-Identifier: Apache-2.0
// Copyright Contributors to the OpenTimelineIO project

#pragma once

#include "nonstd/optional.hpp"
#include "opentimelineio/version.h"

namespace opentimelineio { namespace OPENTIMELINEIO_VERSION {

using nonstd::nullopt;
using nonstd::nullopt_t;
using nonstd::optional;

}} // namespace opentimelineio::OPENTIMELINEIO_VERSION
